module Bidict exposing (Bidict, decoder, empty, encode, getKeys, getValue, insert)

import Dict exposing (Dict)
import Json.Decode as Decode exposing (Decoder)
import Json.Encode as Encode
import Set exposing (Set)


type Bidict key value
    = Bidict
        { values : Dict key value
        , keys : Dict value (Set key)
        }


empty : Bidict key value
empty =
    Bidict
        { values = Dict.empty
        , keys = Dict.empty
        }


getValue : comparable -> Bidict comparable value -> Maybe value
getValue key (Bidict model) =
    Dict.get key model.values


getKeys : comparable -> Bidict key comparable -> List key
getKeys value (Bidict model) =
    Maybe.withDefault [] (Maybe.map Set.toList (Dict.get value model.keys))


insert : comparable -> comparableToo -> Bidict comparable comparableToo -> Bidict comparable comparableToo
insert key value (Bidict model) =
    Bidict
        { model
            | values = Dict.insert key value model.values
            , keys =
                Dict.update value
                    (Just << Set.insert key << Maybe.withDefault Set.empty)
                    model.keys
        }


encode :
    (key -> Encode.Value)
    -> (value -> Encode.Value)
    -> Bidict key value
    -> Encode.Value
encode encodeKey encodeValue (Bidict model) =
    Encode.dict (Encode.encode 0 << encodeValue)
        (Encode.set encodeKey)
        model.keys


decoder : Decoder comparable -> Decoder comparableToo -> Decoder (Bidict comparable comparableToo)
decoder keyDecoder valueDecoder =
    Decode.andThen (decoderHelper valueDecoder empty)
        (Decode.keyValuePairs (Decode.list keyDecoder))


decoderHelper :
    Decoder comparableToo
    -> Bidict comparable comparableToo
    -> List ( String, List comparable )
    -> Decoder (Bidict comparable comparableToo)
decoderHelper valueDecoder (Bidict model) valueAndKeys =
    case valueAndKeys of
        ( valueString, key :: moreKeys ) :: moreKeysAndValues ->
            case Decode.decodeString valueDecoder valueString of
                Ok value ->
                    decoderHelper valueDecoder
                        (Bidict
                            { model
                                | values = Dict.insert key value model.values
                                , keys = Dict.insert value (Set.fromList (key :: moreKeys)) model.keys
                            }
                        )
                        moreKeysAndValues

                Err error ->
                    Decode.fail (Decode.errorToString error)

        [] ->
            Decode.succeed (Bidict model)

        ( valueString, [] ) :: _ ->
            Decode.fail ("The value '" ++ valueString ++ "' has no associated keys. A Bidict must always have at least one key associated with any given value")
